# TODO MultiFrame -> GroupedFrame
#      new MultiFrame for multidata

from __future__ import print_function
from future.builtins import *

import os
# import re
import shelve
from collections import OrderedDict

from pandas import Series, DataFrame, Index
# from pandas import concat

from . import plot as plt
from .plot import style as defstyle
from .plot import compose_styles
from . import io

import bokeh.plotting as bk

import numpy as np

# Series.__repr__ = (lambda x: ("Hash: {}\nTimes: {}\nLoc: {}\nValues: {}".format(
#                     io.hash_series(x),
#                     list(set(x.keys().get_level_values('Time'))), # avoid back and forth conversion
#                     list(set(x.keys().get_level_values('Loc'))),
#                     x.values))) #TODO monkey patch to use hashes
Database = False

if Database:
    case_data_base = shelve.open(os.path.expanduser('~') + "/.owls/db")
else:
    case_data_base = dict()


def read_sets(folder, name="None",
              search=io.FPNUMBER,
              **kwargs):
    def setsfolder(folder):
        p = os.path.join(folder, "postProcessing")
        return (os.path.join(p, "sets") if os.path.exists(p)
                else os.path.join(folder, "sets"))

    return FoamFrame(folder=setsfolder(folder), search_files=False,
                     search_pattern=search, name=name,
                     show_func="plot", preHooks=None,
                     exclude=['processor'], **kwargs)


def read_lag(folder, files, skiplines=1,
             name="None", cloud="[A-Za-z]*Cloud1",
             preHooks=None, decomposed=False, **kwargs):
    search = io.FPNUMBER + "/lagrangian/" + cloud
    search = (search if not decomposed else "processor[0-9]?/" + search)
    return FoamFrame(folder=folder, search_files=files,
                     search_pattern=search, name=name,
                     skiplines=skiplines, show_func="scatter",
                     **kwargs)


def read_eul(folder, files, skiplines=1, name="None",
             decomposed=False, preHooks=None, **kwargs):
    search = io.FPNUMBER
    search = (search if not decomposed else "processor[0-9]?/" + search)
    return FoamFrame(folder=folder, search_files=files,
                     search_pattern=search, name=name,
                     skiplines=skiplines, show_func="scatter",
                     preHooks=preHooks, **kwargs)


def read_exp(folder, name="None", search="", **kwargs):
    return FoamFrame(folder=folder, search_files=False,
                     search_pattern=search, name=name,
                     show_func="scatter", **kwargs)


def read_log(folder, keys, log_name='log', plot_properties=False, name="None"):
    origins, df = io.import_logs(folder, log_name, keys)
    ff = FoamFrame(df)
    ff.properties = Props(
        origins=origins, name=folder,
        plot_properties=plot_properties,
        folder=folder, times=[0], symb="-",
        show_func="plot")
    return ff

""" Filter Helper Functions """

isIn = lambda x: lambda y: x in y
isNotIn = lambda x: lambda y: x not in y


class PlotProperties():

    def __init__(self):
        from collections import defaultdict
        self.properties = defaultdict(dict)

    def insert(self, field, properties):
        self.properties[field].update(properties)
        return self

    def set(self, inserts):
        for k, d in inserts.items():
            self.insert(k, d)

    def select(self, field, prop, default=None):
        field = self.properties[field]
        if not field:
            return
        else:
            return field.get(prop, default)


class Props():
    # TODO default args

    def __init__(self, origins, name,
                 plot_properties, folder,
                 symb, show_func):
        self.origins = origins
        self.name = name
        self.plot_properties = plot_properties
        self.folder = folder
        self.symb = symb
        self.show_func = show_func


class FoamFrame(DataFrame):
    """ Data reprensentation of OpenFOAM field (eulerian and lagrangian)
    and set files. Instantiated through read methods, e.g:
    read_sets, read_lag, read_eul, read_exp


    Examples:
    ----------

    case = read_sets(folder="home/user/case",plot_properties={})
    case.data # access data frame

    Parameters:
    ----------
    folder: data location containing a time or sets folder
    files: search only for files with given name, None for all files
    plot_properties: dictionary for consistent plotting of ranges and ax labels
    skiplines: read only every n-th entry
    cloud: name of lagrangian cloud
    name: case name for plot legends

    Note:
    ----------

    If data is accessed through [] only latest item is returned. For full times
    access iteratetimes() can be used.

    Categories:

        { "rad_pos": lambda field -> position
          "centreLine": [] lambda field -> i of []

        }

        example:
            lambda field: re.search('[0-9]*\.[0-9]*').group()[0]

    TODO:
        use case as cases ojects with a 3-level index
             case['u']
             acces time of all cases -> df.iloc[df.index.isin([1],level=1)]
        refactor plot into case objects itself,
            ?case.show('t','u', time_series = False)
        refactor origins
        make iteratetimes() access a delta
    """
    def __init__(self, *args, **kwargs):

        skip = kwargs.get('skiplines', 1)
        times = kwargs.get('skiptimes', 1)
        name = kwargs.get('name', 'None')
        symb = kwargs.get('symb', 'o')
        files = kwargs.get('search_files', None)
        properties = kwargs.get('properties', None)
        lines = kwargs.get('maxlines', 0)
        search = kwargs.get('search_pattern', io.FPNUMBER)
        folder = kwargs.get('folder', None)
        plot_properties = kwargs.get('plot_properties', PlotProperties())
        show_func = kwargs.get('show_func', None)
        validate = kwargs.get('validate', True)
        preHooks = kwargs.get('preHooks', None)
        exclude = kwargs.get('exclude', [" "])  # FIXME
        times_stride = kwargs.get('times_stride', 1)
        times_range = kwargs.get('times_range', "all") # FIXME implement strides
        times_slice = times_range

        keys = ['skiplines',
                'skiptimes',
                'preHooks',
                'name',
                'symb',
                'search_files',
                'properties',
                'maxlines',
                'search_pattern',
                'folder',
                'plot_properties',
                'show_func',
                'exclude',
                'times_stride',
                'times_range',
                ]

        for k in keys:
            if k in kwargs:
                kwargs.pop(k)

        # TODO explain what happens here
        if folder is None:
            # super(FoamFrame, self).__init__(*args, **kwargs)
            DataFrame.__init__(self, *args, **kwargs)
        else:
            if preHooks:
                for hook in preHooks:
                    hook.execute()
            if (folder in case_data_base) and Database:
                print("re-importing", end=" ")
            else:
                print("importing", end=" ")
            print(name + ": ", end="")
            origins, data = io.import_foam_folder(
                path=folder,
                search=search,
                files=files,
                skiplines=skip,
                maxlines=lines,
                skiptimes=times,
                exclude=exclude,
                times_slice=times_slice
                )
            try:
                DataFrame.__init__(self, data)
            except:
                pass
            self.properties = Props(
                origins,
                name,
                plot_properties,
                folder,
                symb,
                show_func)
            if validate and Database:
                self.validate_origins(folder, origins)
            # register to database
            if Database:
                case_data_base.sync()

    def validate_origins(self, folder, origins):
        origins.update_hashes()
        if case_data_base.has_key(folder):
            if (case_data_base[folder]["hash"] == origins.dct["hash"]):
                print(" [consistent]")
            else:
                entries_new = len(origins.dct.keys())
                entries_old = len(case_data_base[folder].keys())
                if entries_new > entries_old:
                    print("[new timestep] ")
                    # print origins.dct.keys()
                    case_data_base[folder] = origins.dct
                elif entries_new < entries_old:
                    # print folder
                    # print origins.dct.keys()
                    # print case_data_base[folder].keys()
                    print("[missing timestep]")
                    case_data_base[folder] = origins.dct
                elif entries_new == entries_old:
                    print("[corrupted]", end="")
                    for time, loc, field, item in origins.hashes():
                        time_name, time_hash   = time
                        loc_name, loc_hash     = loc
                        field_name, field_hash = field
                        filename, item_hash    = item
                        try:
                            orig_hash = case_data_base[folder][time_name][loc_name][field_name][1]
                        except:
                            orig_hash = item_hash
                        if (item_hash != orig_hash):
                            print("")
                            print("corrupted fields:")
                            print("\t" + field_name + " in " +  filename)
                    case_data_base[folder] = origins.dct
        else:
            case_data_base[folder] = origins.dct

    def source(self, col):
        """ find corresponding file for column """
        # return get time loc  and return dict for every column
        # latest.source['u']
        return

    # ----------------------------------------------------------------------
    # Internal helper methods

    @property
    def _constructor(self):
        # override DataFrames constructor
        # to enable method chaining
        return FoamFrame

    def _is_idx(self, item):
        """ test if item is column or idx """
        itemt = type(item)
        # if item is Series of booleans
        # it cant be an index
        from past.builtins import unicode
        from past.builtins import str as text
        if itemt not in [int, str, float, unicode, text]:
            return False
        else:
            return item in self.index.names


    @property
    def grouped(self):
        return self._is_idx("Group")

    @staticmethod
    def from_dict(input_dict, name="None",
            plot_properties=None, symb=".", show_func="scatter"
            ):
        """ import raw data from python dictionary
            format {(timestep, pos, ): [fields]}
            usage: {(0):[1,2,3]}

        """
        pP = (PlotProperties() if not plot_properties else plot_properties)
        ff = FoamFrame(input_dict, folder=None)
        ff.properties = Props("raw", name, pP, "", symb, show_func)
        return ff

    # ----------------------------------------------------------------------
    # Info methods

    def __str__(self):
        return "FoamFrame: \n" + super(FoamFrame, self).__str__()

    @property
    def times(self):
        """ return times for case """
        return set([_[0] for _ in self.index.values])

    @property
    def locations(self):
        """ return times for case """
        return set([_[1] for _ in self.index.values])

    # ----------------------------------------------------------------------
    # Selection methods

    def __getitem__(self, item):
        """ call pandas DataFrame __getitem__ if item is not
            an index
        """
        if self._is_idx(item):
            try:
                level = self.index.names.index(item)
                return list(zip(*self.index.values))[level]
            except:
                return
                # print("failed ", item) NOTE for debugging
        else:
            if (type(item) is str) and item not in self.columns:
                return Series()
            else:
                return super(FoamFrame, self).__getitem__(item)

    @property
    def latest(self):
        """ return latest time for case """
        ret = self.query('Time == {}'.format(self.latest_time))
        ret.properties = self.properties
        return ret

    @property
    def latest_time(self):
        """ return value of latest time step """
        return max(self.index.levels[0])

    def at(self, idx_name, idx_val):
        """ select from foamframe based on index name and value"""
        # TODO FIX This
        ret = self[self.index.get_level_values(idx_name) == idx_val]
        # match = [(val in idx_val)
        #      for val in self.index.get_level_values(idx_name)]
        # ret = self[match]
        if idx_name == "Group":
            ret.index = ret.index.droplevel("Group")

        ret.properties = self.properties
        return ret

    def id(self, loc):
        """ Return FoamFrame based on location """
        return self.at(idx_name='Pos', idx_val=loc)

    def location(self, loc):
        """ Return FoamFrame based on location """
        return self.at(idx_name='Loc', idx_val=loc)

    def loc_names(self, key):
        """ search for all index names matching keyword"""
        return [_ for _ in  self.index.get_level_values("Loc") if key in _]

    def field_names(self, key):
        """ search for all field names matching keyword"""
        return [_ for _ in self.columns if key in _]

    # ----------------------------------------------------------------------
    # Manipulation methods

    def add(self, data, label):
        """
        Add a given Series

        Usage:
            case.add(sqrt(uu),'u_rms')
        """
        self.latest[label] = data
        return self

    def rename(self, search, replace):
        """ rename field names based on regex """
        import re
        self.columns = [re.sub(search, replace, name) for name in self.columns]

    def rename_idx(self, search, replace):
        """ rename field names based on regex """
        self.index = Index(
            [(t, replace if x == search else x, i) for t, x, i in list(self.index)],
            names=self.index.names)

    def rename_idxs(self, rename_map):
        """ rename multiple field names based dictionary
            of {search: replace} """
        for s, r in rename_map.items():
            self.rename_idx(s, r)

    # ----------------------------------------------------------------------
    # Plotting methods

    def draw(self, x, y, z, title, func, figure, legend_prefix="", titles=None, **kwargs):
        # TODO Rename to _draw
        def _label(axis, field):
            label = kwargs.get(axis + '_label', False)
            if label:
                self.properties.plot_properties.insert(
                    field, {axis + '_label': label})
            else:
                label = self.properties.plot_properties.select(
                    field, axis + '_label', "None")
            return label

        def _range(axis, field):
            from bokeh.models import Range1d
            p_range_args = kwargs.get(axis + '_range', False)
            if p_range_args:
                self.properties.plot_properties.insert(
                    field, {axis + '_range': p_range})
            else:
                p_range = self.properties.plot_properties.select(
                    field, axis + '_range')
            if not p_range:
                return False
            else:
                return Range1d(start=p_range[0], end=p_range[1])

        figure_properties = {"title": title}

        if kwargs.get('x_range', False):
            figure_properties.update({"x_range": kwargs.get('x_range')})
        figure.set(**figure_properties)

        if func == "quad":
            getattr(figure, func)(top=y, bottom=0, left=x[:-1], right=x[1:],
                                  **kwargs)
            return figure

        colors = plt.next_color()
        spec_color = kwargs.get("color", False)
        spec_legend = kwargs.get("legend", False)
        y = (y if isinstance(y, list) else [y])
        for yi in y:
            x_data, y_data = self[x], self[yi]
            # TODO FIXME
            for k in ['symbols', 'order', 'colors', 'symbol']:
                if k in kwargs.keys():
                    kwargs.pop(k)
            if not spec_color:
                kwargs.update({"color": next(colors)})
            if not spec_legend:
                # NOTE title overrides legend, does that make sense always?
                yi = (yi if not title else "")
                if yi and legend_prefix:
                    legend = legend_prefix + "-" + yi
                if not yi and legend_prefix:
                    legend = legend_prefix
                if not legend_prefix:
                    legend = yi

                kwargs.update({"legend": legend})

            getattr(figure, func)(x=x_data,
                                  y=y_data,
                                  **kwargs)

        for ax, data in {'x': x, 'y': y[0]}.items():
            if _label(ax, data):
                getattr(figure, ax+'axis')[0].axis_label = _label(ax, data)
            # setattr(getattr(figure, ax + 'axis'),
            #         'axis_label', _label(ax, data))
            if _range(ax, data):
                r = setattr(figure, ax+'_range', _range(ax, data))
        return figure

    def histogram(self, y, x=None, title="", figure=False, **kwargs):
        figure = (figure if figure else plt.figure())
        import numpy as np
        hist, edges = np.histogram(self[y], density=True, bins=50)

        return self.draw(x=edges, y=hist, z=None, title=title, func="quad", figure=figure, **kwargs)

    def scatter(self, y, x='Pos', z=False, title="", figure=False, **kwargs):
        figure = (figure if figure else plt.figure())
        return self.draw(x, y, z, title, func="scatter", figure=figure, **kwargs)

    def plot(self, y, x='Pos', z=False, title="", figure=False, **kwargs):
        figure = (figure if figure else plt.figure())
        if kwargs.get('symbol', None):
            kwargs.pop('symbol')
        return self.draw(x, y, z, title, func="line", figure=figure, **kwargs)


    def show(self, y, x="Pos", figure=False,
             overlay="Field", style=defstyle,
             legend_prefix="", post_pone_style=False,
             row=None, titles=None, **kwargs):
        style = (compose_styles(style, []) if isinstance(style, list) else style)
        if kwargs.get("props", False):
            props = kwargs.pop("props")
            self.properties.plot_properties.set(props)

        def create_figure(y_, f, title=""):
            if kwargs.get("title"):
                title = kwargs.get("title")
                kwargs.pop("title")
            return getattr(self, self.properties.show_func)(y=y_, x=x, figure=f, legend_prefix=legend_prefix, title=title, **kwargs)

        def create_figure_row(y, arow=None):
            arow = (arow if arow else OrderedDict())
            if not self.grouped:
                y = (y if isinstance(y, list) else [y])
                if overlay == "Field":
                    # SINGLE FIGURE MUTLIPLE FIELDS
                    fig_id, f = (figure if figure else ("".join(y), plt.figure()))
                    for yi in y:
                        create_figure(y, f)
                    arow[fig_id] = f
                if not overlay:
                    # MULTIPLE FIGURES
                    # test if figure with same id already exists
                    # so that we can plot into it
                    # otherwise create a new figure
                    for i, yi in enumerate(y):
                        title = ("" if not titles else titles[i])
                        fig_id, f = ((yi, arow[yi])
                                     if arow.get(yi, False)
                                     else (yi, plt.figure(title=title)))
                        arow[fig_id] = create_figure(yi, f, title=title)
            if self.grouped:
                groups = list(set(self["Group"]) if self["Group"] else set())
                groups.sort()
                if overlay == "Group":
                    # ALIGN ALONG GROUPS
                    # for every yi a new figure is needed
                    #arow = (arow if arow else OrderedDict())
                    for yi in y:
                        fig_id, f = ((yi, arow[yi])
                                if arow.get(yi, False)
                                else (yi, plt.figure()))
                        colors = plt.next_color()
                        for name in groups:
                            color = next(colors)
                            arow[yi] = self.at("Group", name).show(
                                    x=x, y=yi, title=yi, figure=(fig_id, f),
                                    post_pone_style=True, overlay="Field",
                                    color=color, legend_prefix=legend_prefix,
                                    legend=str(name),
                                    **kwargs)[yi]
                if overlay == "Field":
                    # row = (arow if arow else OrderedDict())
                    for group in groups:
                        fig_id, f = ((group, arow[group])
                                     if arow.get(group, False)
                                     else (group, plt.figure()))
                        field = self.at("Group", group)
                        arow[group] = field.show(x=x, y=y, title=str(group),
                                                 figure=(group, f), overlay="Field",
                                                 post_pone_style=True,
                                                 legend_prefix=legend_prefix,
                                                 **kwargs)[group]
            return arow

        fig_row = create_figure_row(y, row)
        return (fig_row
                if post_pone_style
                else bk.GridPlot(children=style(rows=[list(fig_row.values())]))
               )


    def show_func(self, value):
        """ set the default plot style
            valid arguments: scatter, plot """
        self.properties.show_func = value

    def set_plot_properties(self, **values):
        """ set plot properties  """
        self.properties.plot_properties.set(values)

    # ----------------------------------------------------------------------
    # Filter methods

    def filter_fields(self, name, lower, upper):
        """ filter based on field values

            Examples:

                .filter_fields('T', 1000, 2000)
        """
        return self.filter(name, field=lambda x: lower < x < upper)

    def filter_locations(self, index):
        """ filter based on locations

            Examples:

                .filter_location(Owls.isIn('radial'))
                .filter_location(Owls.isNotIn('radial'))

        """
        return self.filter(name='Loc', index=index)

    def filter(self, name, index=None, field=None):
        """ filter on index or field values by given functioni

            Examples:

                .filter(name='T', field=lambda x: 1000<x<2000)
                .filter(name='Loc', index=lambda x: 0.2<field_to_float(x)<0.8)
        """
        if index:
            ret = self[list(map(index, self.index.get_level_values(name)))]
            ret.properties = self.properties
            return ret
        elif field:
            ret = self[list(map(field,self[name]))]
            ret.properties = self.properties
            return ret
        else:
            return self

    # ----------------------------------------------------------------------
    # Grouping methods

    def by_index(self, field, func=None):
        func = (func if func else lambda x: x)
        return self.by(field, func)

    def by_field(self, field, func=None):
        func = (func if func else lambda x: x)
        return self.by(field, func)

    def by_location(self, func=None):
        func = (func if func else lambda x: x)
        return self.by("Loc", func)

    def by_time(self, func=None):
        func = (func if func else lambda x: x)
        return self.by("Time", func)

    def by(self, name, func):
        ret = self.copy() # Too expensive ? pd.concat( [A, pd.DataFrame(s)], axis=1 )
        ret.properties = self.properties
        if self._is_idx(name):
            index_values = ret.index.get_level_values(name)
            ret["Group"] = index_values.map(func)
        else:
            ret["Group"] = ret[name].map(func)
        ret.set_index("Group", append=True, inplace=True)
        ret.reorder_levels(['Time', 'Loc', 'Pos', 'Group'])
        return ret



# for snappy
import subprocess, os, json
from IPython.display import HTML, display
from glob import glob
import shutil

import numpy as np

TOOLS = "pan,wheel_zoom,box_zoom,reset,save"

colored = ["black", "blue", "fuchsia", "gray", "green",
           "lime", "maroon", "navy", "olive", "orange", "purple",
           "red", "silver", "teal", "yellow", "aqua"]

symbols_full = ["circle", "square", "triangle", "diamond", "inverted_triangle"]

white2black = [
    '#000000', '#0D0D0D', '#1A1A1A', '#262626', '#333333',
    '#404040', '#4C4C4C', '#595959', '#666666', '#737373',
    '#808080', '#8C8C8C', '#999999', '#A6A6A6', '#B2B2B2',
    '#BFBFBF', '#CCCCCC', '#D9D9D9', '#E6E6E6', '#F2F2F2',
    '#FFFFFF'][::-1]

config = {
    "color_cycle": colored,
    "symbol_cycle": symbols_full,
    }


def head(elems):
    return [elems[0]]


def tail(elems):
    return elems[1:]


def all(elems):
    return elems


def rtail(elems):
    return elems[:-1]


def last(elems):
    return [elems[-1]]


# def adjustRow(style, whereRow, whereFigs=None, rows=None):
#     """ adjusts a list of figures to given geometry """
#     for f in where(figs):
#         for key, value in style.items():
#             if '.' in key:
#                 _ = key.split('.')
#                 f = getattr(f, _[0])
#                 key = _[1]
#             setattr(f, key, value)
#     return figs

def adjustColumn(style, whereRow, whereFigs=None, rows=None):
    for row in whereRow(rows):
        row = (whereFigs(row) if whereFigs else row)
        for fig in row:
            for key, value in style.items():
                if '.' in key:
                    _ = key.split('.')
                    fig = getattr(fig, _[0])
                    key = _[1]
                setattr(fig, key, value)
    return rows


def figure(**kwargs):
    import bokeh.plotting as bk
    return bk.figure(tools=TOOLS, **kwargs)


# def merge(*args, **kwargs):
#     import bokeh.plotting as bk
#     figure = (bk.figure() if not kwargs.get('figure', False) else
#               kwargs.get('figure'))
#     y = kwargs.get('y', None)
#     x = kwargs.get('x', 'Pos')
#     try:
#         kwargs.pop('y')
#         kwargs.pop('x')
#     except:
#         pass
#     y = (y if type(y) == list else [y]*len(args))  # FIXME do the same for x
#     kwargs.pop('figure')
#     legend = kwargs.get('legend')
#     if legend:
#         kwargs.pop('legend')
#     for yi, p in zip(y, args):
#         if legend:
#             kwargs.update({'legend': p.properties.name})
#         override_color = p.properties.plot_properties.properties.get('Color', False)
#         color = (override_color if override_color else next(kwargs["colors"]))
#         p.show(x=x, y=yi, color=color, symbol=next(kwargs["symbols"]), figure=figure, **kwargs)
#     return figure
#
#
# def multi_merge(*args, **kwargs):
#     """ call merge for all args
#
#         Examples:   mm=multi_merge(
#                         sets1.latest.by_index('Loc'),
#                         sets2.latest.by_index('Loc'),
#                         by='[0-9]+',
#                         x='Pos',
#                         y='vMean'
#                         order=[x-10,x+25])
#
#     """
#     import bokeh.plotting as bk
#     y = kwargs.get('y', None)
#     x = kwargs.get('x', 'Pos')
#     if kwargs.get('legend'):
#         legend = kwargs.get('legend')
#         kwargs.pop('legend')
#     else:
#         legend = False
#         kwargs.pop('legend')
#     plots = []
#     c = args[0]
#     # go through all items to be plotted
#     items = (
#         ((name, data) for name, data in c.items() if name in kwargs['order'])
#         if kwargs.get('order', False) else c.items()
#     )
#     for nr, (name, data) in enumerate(items):
#         sub_plots = [data]
#         colors = next_color()
#         symbols = next_symbol()
#         figure = bk.figure()
#         for c_ in args[1:]:
#             # and through all sets to be plotted
#             for name_, plot_ in c_.items():
#                 if not kwargs.get('order', False):
#                     # select by regex
#                     # now see if we have a match
#                     selector = kwargs.get('by', "[A-Za-z0-9_\-]")
#                     # skip if search is empty
#                     if (not re.search(selector, name) or
#                         not re.search(selector, name_)):
#                         continue
#                     # append to subplot if same schema
#                     if (re.search(selector, name).group()
#                         == re.search(selector, name_).group()):
#                         sub_plots.append(plot_)
#                 else:
#                     # select by name in order list
#                     if name_ == name:
#                         sub_plots.append(plot_)
#         for kw in ['x', 'y']:
#             try:
#                 kwargs.pop(kw)
#             except:
#                 pass
#         if kwargs.get('legend_pos', 0) != nr:
#             legend = False
#         title = (kwargs.get('titles')[nr] if kwargs.get('titles', False) else name)
#         plots.append(merge(
#             *sub_plots, x=x, y=y,
#             title=title, colors=colors,
#             symbols=symbols, figure=figure,
#             legend=legend, **kwargs))
#     return plots


def plot_cases(cases, y, order, x='Pos', legend=True, **kwargs):
    from .FoamFrame import FoamFrame
    """ plot all cases in cases dict at specified locations
        and latest time step """
    elems = [x.by_index('Loc') for x in cases.values() if type(x) is FoamFrame]
    return multi_merge(*elems, x=x, y=y, order=order, legend=legend, **kwargs)


def next_color():
    from itertools import cycle
    for col in cycle(config['color_cycle']):
        yield col


def next_symbol():
    from itertools import cycle
    for sym in cycle(config['symbol_cycle']):
        yield sym

# Themes

from functools import partial
from functools import reduce as freduce
from functional import compose


rcParams = {
    'fig_size': {'min_border': 12,
                 'plot_height': 350,
                 'plot_width': 300},
    'font': {'axis.axis_label_text_font_size': "14pt",
             #'legend.label_text_font_size': "10pt",
             #'axis.major_label_text_font_style': 'bold',
             #'axis.axis_label_text_font_style': 'bold'
             },
    'color': {"outline_line_color": "black"}
    }

multi_compose = partial(freduce, compose)
compose_styles = lambda x, y: multi_compose(x + y)

#axis_label_text_font_style, major_label_text_font_style or axis_label_text_font_size
#'min_border':12, 'plot_height': 350, 'plot_width': 350,
rowLabel = partial(
    adjustColumn,
    {'yaxis.visible': False},
    all, tail)

size = partial(adjustColumn, rcParams['fig_size'], all, all)
font = partial(adjustColumn, rcParams['font'], all, all)
color = partial(adjustColumn, rcParams['color'], all, all)

cleanTitle = partial(adjustColumn, {'title': ""}, tail, all)
cleanXAxis = partial(adjustColumn, {'xaxis.visible': False}, rtail, all)
cleanYAxis = partial(adjustColumn, {'yaxis.visible': False}, all, tail)


# reset legends
# all but last
cleanLegendIa = partial(adjustColumn, {'legend.legends': []}, rtail, all)
cleanLegendIb = partial(adjustColumn, {'legend.legends': []}, all, rtail)
cleanLegendIIa = partial(adjustColumn, {'legend.border_line_color': None}, all, rtail)
cleanLegendIIb = partial(adjustColumn, {'legend.border_line_color': None}, rtail, all)
cleanLegenda = partial(adjustColumn, {'legend.legends': []}, all, all)
cleanLegendb = partial(adjustColumn, {'legend.border_line_color': None}, all, all)
# all but first
cleanLegendIc = partial(adjustColumn, {'legend.legends': []}, all, tail)
cleanLegendId = partial(adjustColumn, {'legend.legends': []}, tail, all)
cleanLegendIIc = partial(adjustColumn, {'legend.border_line_color': None}, all, tail)
cleanLegendIId = partial(adjustColumn, {'legend.border_line_color': None}, tail, all)
cleanLegendb = partial(adjustColumn, {'legend.border_line_color': None}, all, all)

def legend(pos):
    return partial(adjustColumn, {'legend.orientation': pos}, all, all)

lastLegend = [cleanLegendIa, cleanLegendIb, cleanLegendIIa, cleanLegendIIb]
firstLegend = [cleanLegendIc, cleanLegendId, cleanLegendIIc, cleanLegendIId]
noLegend = [cleanLegenda, cleanLegendb]

legendTopLeft = legend("top_left")
legendBottomLeft = legend("bottom_left")
legendTopRight = legend("top_right")
legendBottomRight = legend("bottom_right")


lastLegend = [cleanLegendIa, cleanLegendIb, cleanLegendIIa, cleanLegendIIb]


arangement = lambda x: np.array(x).reshape(greatest_divisor(len(x)),-1).tolist()

default_style = [size, font, color]

def greatest_divisor(number):
    if number == 1:
        return 1
    for i in reversed(range(number)):
        if number % i == 0:
            return i
    else:
        return 1

style = multi_compose(default_style)
# snapshots


class Snappy():

    def __init__(self, path, config, out_path):
        self.path = path
        self.baseName = os.path.basename(self.path)
        self.config = config
        self.out_path = out_path
        self.snappy_path = os.path.dirname(__file__) + "/../scripts/snap.py"

    def generate(self, animate=False):
        old_dir = os.getcwd()
        os.chdir(self.path)
        args = " -d " if not animate else " -da --gif "
        cmd = "python {} {} --json_config=\'{}\' &".format(
              self.snappy_path, args, self.config)
        subprocess.check_call(cmd, shell=True)
        os.chdir(old_dir)

    def generate_image_list(self, animate=False, ext=".mp4"):
        images = []
        vals = json.loads(self.config)
        def appendr(images, prefix, extension):
            for s in vals["slices"].keys():
                for f in ["vectors", "scalars"]:
                    for v in vals[f].keys():
                        iname = self.baseName + "/" + prefix + s + v + extension
                        images.append(iname)
            return images

        if not animate:
            images = appendr(images, "last_", "_0.png")
            images = appendr(images, "last_", "_1.png")
            images = appendr(images, "last_", "_2.png")
        else:
            images = appendr(images, "", ext)
        self.images=images
        return images

    def copy_to_notebook(self):
        for p in [self.out_path, os.path.join(self.out_path, self.baseName)]:
            if not os.path.exists(p): os.makedirs(p)
        for ext in ["*.png", "*.mp4", "*.gif"]:
            for fn in glob(os.path.join(self.path + "/postProcessing/anim", ext)):
                    shutil.copy(fn, os.path.join(self.out_path, self.baseName))

    def display(self, size, media_type="mp4"):
        if media_type == "mp4":
            i = '<video width={} height="222" controls="controls"> <source src={} type="video/mp4" /></video>'
        else:
            i = "<img style='width: {}px; margin: 0px; float: left; border: 1px solid black;' src='{}' />"
        images = ''.join([i.format(size, s) for s in self.images])
        return display(HTML(images))



import numpy as np
from . import io


class Mesh():

    def __init__(self, path):
        self.path = path + "/constant/polyMesh/"
        print("reading mesh", end="")
        self.df = io.import_foam_mesh(path)
        print(" [done]\ncomputing face centres ", end="")
        self.faceCtr = self.compute_faceCentres()
        print("[done]\ncomputing cell centres ", end="")
        self.cellCtrs = self.compute_centres()
        print("[done]")

    def clean(self, name):
        return (self.df[self.path + name].reset_index(level=['Loc','Id'], drop=True))

    @property
    def owner(self):
        o = "owner"
        return self.clean(o)[o]

    @property
    def neighbour(self):
        n = "neighbour"
        return self.clean(n)[n]

    @property
    def faces(self):
        return self.clean("faces")

    @property
    def points(self):
        return self.clean("points")


    def compute_faceCentres(self):
        faceCtrsX = lambda x: self.points.ix[self.faces[x].values.astype(int)].reset_index(drop=True)
        faceCtrs = faceCtrsX("faces_0")
        for i, f in enumerate(self.faces):
            if i == 0:
                continue
            faceCtrs = faceCtrs.add(faceCtrsX(f), fill_value=0.0)
        return faceCtrs/4.0

    def compute_centres(self):
        """ compute cell centres for given mesh """
        # As in primitiveMesh.H we compute an approx.
        # cell centre first as the average of the face centres
        o = self.faceCtr.set_index(mesh.owner).sort_index()
        n = self.faceCtr[0:len(self.neighbour)].set_index(self.neighbour).sort_index()
        return o.append(n).groupby(level=0).mean()

    def line(self, start, end):
        pass



from __future__  import print_function
from future.builtins import *

import matplotlib.pyplot as plt
import matplotlib as mpl

from pandas import Series

class Process():

    def __init__(self):
        pass
    
    def execute(self,x,y):
        t0 = x[0]
        ave_els = [y[0]]
        for i,t in enumerate(x[1:]):
            dt = t - t0
            rt = 1/t
            alpha = t0*rt
            beta  = dt*rt
            ave_els.append(alpha*ave_els[-1] + beta*y[i+1])
            t0 = t
        average = Series(ave_els)
        average.name = "avg_" + y.name
        return {'y':average, 'x':x, 'symbol':'--'}

class Annotation():

    def __init__(self, x, y, text, subplot):
        self.text = text
        self.subplot = subplot
        self.x = x
        self.y = y

class Plot():
    ''' wrapper class around mpl
        supported styles: [classic, jet, wide, dcol, row]
    '''

    def __init__(self, geom=[1,1], leg=[0],
            style='classic',
            cmap=mpl.cm.gray,
            case="",
            figurename="figure",
            backupdir="",
            backup=False,
            fileformat="eps",
            ):

        self.nx, self.ny = ((1,1) if not geom else (geom[0], geom[1]))
        self.leg = leg
        self.cmap = cmap
        self.style = style
        self.plots = []
        self.origins = []
        self.backupdir = backupdir
        self.case_dir = backupdir + case + "/"
        self.sets_dir = self.case_dir + "sets/"
        self.data_dir = self.sets_dir + figurename
        self.figurename = "{}{}.{}".format(self.case_dir,
                                        figurename,
                                        fileformat)
        time = datetime.datetime.now().strftime("%Y-%m-%d_%H:%M")
        self.figurename_date = "{}{}_at_{}.{}".format(self.case_dir,
                                        figurename,
                                        time,
                                        fileformat)
        self.backup = backup
        self._create_backup_dir()
        self.annotations =[]

    def _create_backup_dir(self):
        try:
            os.makedirs(self.data_dir)
        except:
            pass

    def _md5sum(self, filename):
        """ compute md5sum of origin """
        md5 = hashlib.md5()
        with open(filename,'rb') as f: 
            for chunk in iter(lambda: f.read(128*md5.block_size), b''): 
                 md5.update(chunk)
        return md5.hexdigest()[:8]

    def objects(self):
        """ find all plot objects in self.plots """
        for p in self.plots:
            for name, item in p.iteritems():
                if name in ('x','y','z'):
                    yield item

    def backup_origin(self):
        """ copy origins to backup dest"""
        for obj in self.objects():
            try:
                origin = obj.origin
                file_hash = self._md5sum(origin)
                file_dest = "{}/{}_{}_{}".format(self.data_dir,
                        origin.split('/')[-1],
                        obj.name,
                        file_hash)
                shutil.copy2(origin, file_dest)
            except Exception as e:
                pass #print "back up of origins failed: " + str(e) 

    def _savefig_and_origins(self):
        self.backup_origin()
        self.f.savefig(self.figurename)
        self.f.savefig(self.figurename_date)


    def add(self,
          **kwargs
        ):
        """ add series for plotting, valid arguments are 
            x,y,z = Series
            name  = name of the series to displayed in the legend box, default is y.name
            symbol 
            func = 
        """
        try:
            if len(kwargs['x']) == len(kwargs['y']) and len(kwargs['x']) != 0:
                if kwargs.get('func', False):
                    self.plots.append(kwargs['func'].execute(
                                                        kwargs['x'],
                                                        kwargs['y']))
                self.plots.append(kwargs)
        except Exception as e:
            print(e)
            pass

    def create_plot_array(self):
        height = 4
        width = (2*height if self.style == "wide" else height)
        width = (4*height if self.style == "extra_wide" else height)
        return  plt.subplots(self.ny, self.nx,
                    figsize=(self.nx*width, self.ny*height)
                    )

    def first_col(self, index):
        ''' return true if given subplot index is in first col'''
        return (True if (index) % self.nx == 0 else False)

    def last_col(self, index):
        ''' return true if given subplot index is in last col'''
        return (True if (index+1) % self.nx == 0 else False)

    def last_row(self, index):
        ''' return true if given subplot index is in last row'''
        return (True if index >=  self.nx * (self.ny-1) else False)

    def decorate_x(self, index):
        ''' returns true when ax is in last row or style is matrix '''
        if self.style == 'matrix':
            return True
        else:
            return (True if self.last_row(index) else False)

    def decorate_y(self, index):
        ''' returns true if ax is in first col or style is matrix '''
        if self.style == 'matrix':
            return True
        elif self.style in ('jet', 'dcol'):
            return self.first_col(index) or self.last_col(index)
        else:
            return self.first_col(index)

    def invert_x(self, index):
        ''' returns true for jet style if plot is on left side'''
        if self.style == 'jet':
            return self.first_col(index)
        else:
            return False

    def mirror_label(self, index):
        ''' true if in jet and dcol mode and on last col '''
        if self.style in ('jet', 'dcol') :
            return self.last_col(index)
        else:
            return False

    def add_legend(self, index):
        ''' adds a legend if requested '''
        return (True if index in self.leg else False)


    def ax_labels(self):
        ''' create a list of ax labels for every plot '''
        self.xlabels = ['no label' for i in range(self.nx*self.ny)]
        self.ylabels = ['no label' for i in range(self.nx*self.ny)]
        self.xranges = ["auto" for i in range(self.nx*self.ny)]
        self.yranges = ["auto" for i in range(self.nx*self.ny)]
        for plot in self.plots:
            plot_index = plot.get('subplot', 0)
            try:
                self.xlabels[plot_index] = plot['x'].label
                self.xranges[plot_index] = plot['x'].data_range
            except:
                 self.xlabels[plot_index] = "None"
            try:
                self.ylabels[plot_index] = plot['y'].label
                self.yranges[plot_index] = plot['y'].data_range
            except:
                 self.ylabels[plot_index] = "None"

    def iterate_axis(self):
        if type(self.ax) != numpy.ndarray:
            yield 0, self.ax[0]
        else:
            axs = self.ax.flatten()
            for i, a in enumerate(axs):
                yield i, a

    def decorate_axis(self):
        st = self.style
        for index, a in self.iterate_axis():
            self.add_legend(index)
            if self.decorate_y(index):
                a.set_ylabel(self.ylabels[index]) # move to decorate func
                if self.mirror_label(index):
                    a.yaxis.set_ticks_position("right")
                    a.yaxis.set_label_position("right")
            else:
                a.set_yticklabels([])
            if self.decorate_x(index):
                a.set_xlabel(self.xlabels[index]) # move to decorate func
            else:
                a.set_xticklabels([])
            x_range = self.xranges[index]
            if x_range != "auto" and  x_range != "auto":
                if self.invert_x(index):
                    a.set_xlim(x_range[1], x_range[0])
                else:
                    a.set_xlim(x_range[0], x_range[1])
                y_range = self.yranges[index]
                a.set_ylim(y_range[0], y_range[1])
            a.locator_params(nbins=10)
            a.grid(True)

    def draw(self, plot):
        style = plot.get('plot_type', 'scatter')
        x = plot['x']
        y = plot['y']
        z = plot.get('z', False)
        l = (plot.get('name') if plot.get('name',False) else plot['y'].name)
        i = plot.get('subplot', 0)
        s = plot.get('symbol', '.')
        c = plot.get('color', False)
        self.labels.append(l)

        if s in ['o', '.', ',', '+', 'x' ]:
            style = "scatter"
        else:
            style = "plot"

        if style == 'scatter' and type(z) is not bool:
            return self.ax[i].scatter(x, y, c=z, s=40, cmap=self.cmap,
                                     marker='.', lw=0, label=l),

        elif style == 'scatter' and not z:
            return self.ax[i].scatter(x, y, s=40, marker=s,
                                 lw=2, label=l, facecolors='none'),

        if not style == 'scatter':
            if c:
                return self.ax[i].plot(x, y, label=l, lw=3, linestyle=s, color=c),
            else:
                return self.ax[i].plot(x, y, label=l, lw=3, linestyle=s),

    def show(self, suppress=False):
        self.f, ax = self.create_plot_array()
        try:
            self.ax = ax.flatten()
        except:
            self.ax = [ax]
        self.ax_labels()
        self.decorate_axis()
        self.lines = []
        self.labels = []
        for plot in self.plots:
            self.lines.append(self.draw(plot))
        for i, ax in self.iterate_axis():
            if self.add_legend(i):
                ncol = (1 if len(self.plots)/len(self.ax) < 3 else 2)
                ax.legend(ncol=ncol)
        for ann in self.annotations:
            subplot = ann.subplot
            posx = (ann.x if ann.x else max(self.ax[subplot].get_xlim())*0.2)
            posy = (ann.y if ann.y else max(self.ax[subplot].get_ylim())*0.8)
            pos = (posx,posy)
            self.ax[ann.subplot].annotate(ann.text, xy=pos, xytext=pos)
        if self.backupdir and self.backup:
            self._savefig_and_origins()
        if suppress:
            plt.close()

    def zoom(self, subplot=0, x=None, y=None):
        if x:
            self.ax[subplot].set_xlim(x[0],x[1])
        if y:
            self.ax[subplot].set_ylim(y[0],y[1])

    def annotate(self, subplot=0, text="", position="ul", x=False, y=False):
        """ add an annotation to a subplot """
        self.annotations.append(
                Annotation(text=text, subplot=subplot, x=x, y=y)
            )

    def pci_style(self):
        plt.rcParams.update({'font.size': 14, 'family': 'Helvetica'})
        plt.rcParams.update({'weight': 'bold'})
        plt.rcParams.update({'axes.linewidth': 2.0})
        plt.rcParams.update({'axes.titlesize': 'medium'})
        plt.rcParams.update({'figure.subplot.hspace': 0.1,})
        plt.rcParams.update({'figure.subplot.wspace': 0.1,})
        plt.rcParams.update({'axes.labelweight': 'medium'})
        plt.rcParams.update({'legend.scatterpoints': 1})
        plt.rcParams.update({'legend.numpoints': 1})
        plt.rcParams.update({'axes.color_cycle': ['k']})

    def color_style(self):
        plt.rcParams.update({'font.size': 14, 'family': 'Helvetica'})
        plt.rcParams.update({'weight': 'bold'})
        plt.rcParams.update({'axes.linewidth': 2.0})
        plt.rcParams.update({'axes.titlesize': 'medium'})
        plt.rcParams.update({'figure.subplot.hspace': 0.1,})
        plt.rcParams.update({'figure.subplot.wspace': 0.1,})
        plt.rcParams.update({'axes.labelweight': 'medium'})
        plt.rcParams.update({'legend.scatterpoints': 1})
        plt.rcParams.update({'legend.numpoints': 1})
        plt.rcParams.update({'axes.color_cycle': ['k','r', 'g', 'b', 'y','c', 'm']})

def intToColor(i, itot=4):
    cmap = mpl.cm.jet  # I set colomap to 'jet'
    norm = mpl.colors.Normalize(vmin=0, vmax=itot)
    return cmap(norm(i))


def add_colorbar(data, cs, f):
        cb = f.colorbar(cs, pad=0.05)
        cb.set_label(data.get('z_label', 'None'))


def makeHistogram(Series, nbins=50, limits=[], labels=[], legends=[]):
    plt.rcParams.update({'font.size': 18})
    plt.rcParams.update({'family': 'Helvetica'})
    plt.rcParams.update({'weight': 'bold'})

    f, ax = plt.subplots(1, 1, figsize=(15, 6))
    for i, serie in enumerate(Series):
        count, bins = np.histogram(serie, bins=nbins)
        cs = ax.plot(bins, count)
    ax.set_xlim(lim[0], lim[1])
    ax.set_ylim(lim[2], lim[3])
    ax.set_xlabel(title[0])
    ax.set_ylabel(title[1])
    ax.locator_params(nbins=8)
    return f, ax

def get_data(data, time, log, x):
    data_type = x[4]
    base_file = x[2]
    data_loc = x[3]
    if log:
        try:
            d = data[data_type][base_file]
            return d
        except:
            print("trying to acess {} {}".format(base_file, data_loc))
            return 0
    else:
        d = data[data_type][time]['{}_{}'.format(base_file, data_loc)]
        return d


def plot_dict_histogram(data, name, x, legend="", time=False, nbins=100):
    set_ = x[4]
    if not time:
        time = str(max(data[name][set_].keys()))

    id_ = name + legend

    data_ = '{}_{}'.format(x[2],x[3])
    count, bins = np.histogram(data[name][set_][time][data_],
                               bins=nbins, density=True,
                              )

    dict_ = {id_: {'x': fenceAve(bins),
                   'y': count,
                   'x_label': x[0],
                   'y_label': 'density',
                   'x_lim': x[1],
                   'y_lim': [0,1.1*max(count)],
                   }}
    dict_[id_]['plot_type'] = "plot"
    return dict_

def sort_by_pos(data, name, x, y, pos):
    setx = x[4]
    time = str(max(data[name][setx].keys()))
    datax = '{}_{}'.format(x[2],x[3])
    datay = '{}_{}'.format(y[2],y[3])
    data_m = '{}_-_{}'.format(x[2],x[3])
    data_p = '{}_+_{}'.format(x[2],x[3])

    data[name][setx][time][data_m] = data[name][setx][time][datax][ data[name][setx][time][datay] < pos]
    data[name][setx][time][data_p] = data[name][setx][time][datax][ data[name][setx][time][datay] > pos]

def calc_diff(name, x, y):
    setx = x[4]
    time = str(max(name[setx].keys()))
    datax = '{}_{}'.format(x[2],x[3])
    datay = '{}_{}'.format(y[2],y[3])
    data_diff = '{}diff{}_{}'.format(x[2],y[2],x[3])

    name[setx][time][data_diff] = name[setx][time][datax] - name[setx][time][datay]


def plot_dict_rhistogram(data, name, x, legend="", time=False, nbins=100):
    set_ = x[4]
    if not time:
        time = str(max(data[name][set_].keys()))

    id_ = name + legend

    data_ = '{}_{}'.format(x[2],x[3])
    count, bins = np.histogram(data[name][set_][time][data_],
                               bins=nbins, density=True,
                              )

    dict_ = {id_: {'y': fenceAve(bins),
                   'x': count,
                   'y_label': x[0],
                   'x_label': 'density',
                   'y_lim': x[1],
                   'x_lim': [0,1.1*max(count)],
                   }}
    dict_[id_]['plot_type'] = "plot"
    return dict_



from __future__  import print_function
import os

owls_path = os.path.expanduser('~') + "/.owls"
if not os.path.exists(owls_path):
    os.makedirs(owls_path)

from .version import __version__
# from . import io
from .plot import *
# from .foam import *
from .FoamFrame import *
from .MultiFrame import *

from future.builtins import *
print("Owls Version: " + __version__)

try:
  from mplinterface import *
except:
  print("Warning no matplotlib support")



#!/usr/bin/ipython
'''
Reads and converts OpenFoam logfiles and data
to Pandas DataFrames and Series


Bad Karma:
    * read_data_file returns field names which is redundant to data.columns
    * rename to FoamAna.io

'''
from __future__  import print_function
from future.builtins import map, filter, open, next, range, zip, dict, str

import numpy as np
import re
import os
import hashlib
from pandas import DataFrame, concat
from collections import defaultdict, OrderedDict
from IPython.display import display, clear_output

FPNUMBER = "[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?"

FOAM_HEADER = """
/*--------------------------------*- C++ -*----------------------------------*\\
| =========                 |                                                 |
| \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox           |
|  \\    /   O peration     | Version:  2.0.x                                 |
|   \\  /    A nd           | Web:      www.OpenFOAM.org                      |
|    \\/     M anipulation  |                                                 |
\\*---------------------------------------------------------------------------*/
FoamFile
{{
    version     2.0;
    format      ascii;
    class       vol{}Field;
    location    {};
    object      {};
}}
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

dimensions      [0 0 0 0 0 0 0];

internalField   nonuniform List<{}>
"""

DEBUG = True

####################### helper functions ################################
def match(d, event):
    """ returns d.item if reg_exp_key matches event """
    for reg_exp_key in d.keys():
        if re.match(reg_exp_key, event):
            return d[reg_exp_key]

def find_datafiles(
        path=False,
        files=False,
        search=FPNUMBER,
        exclude=None,
        times_slice=None,
    ):
    """ Find all datafiles in each time folder,

        Returns a dictionary of lists containing data
        files for every found time step

        fold: list of time folders to look for data files
              if False time folders in cwd will be taken
        filelist: A list of file names which are accepted,
                  if false all files will be returned
        subfolder: specify wheter to search in cwd or in a specific subfolder
                   accepting a search pattern
        Returns:
            Ordered dict with times as key and
            list of found files
    """
    data_folders = find_datafolders(search, path, exclude, times_slice)
    return OrderedDict(
        [(time, _get_datafiles_from_dir(time, files))
            for time in data_folders]
        )


def find_datafolders(regex, path=False, exclude=None, slice_=None):
    """ Find data folders according to regex
        replaces old find_times function
        Returns sorted list of times as strings """
    path = (path if path else os.getcwd())
    if not path.endswith('/'):
        path = path + '/'
    complete_regex = path + regex + "$"
    folders = []
    for fold, dirs, _ in os.walk(path):
        if exclude:
            dirs[:] = [d for d in dirs for ex in exclude if not re.match(ex, d)]
        folders.append(fold)

    folders = [_ for _ in folders if re.match(complete_regex, _)]
    folders.sort()
    # if slice_ == "all":
    #     return folders
    if slice_ == "latest":
        return folders[-1]
    else:
        return folders
    # FIXME raise Exception
    # return (folders if not slice_ else folders(slice_))


def _get_datafiles_from_dir(path=False, fn_filter=False):
    """ Return file names of Foam files from cwd if no path
        is specified explicitly.
        If no filter list is given the complete list of files will be returned
        else only files matching that list
    """
    path = (path if path else os.getcwd() + "/")
    path = (path + "/" if not path.endswith("/") else path)
    cur_dir = os.walk(path)
    root, dirs, files = next(cur_dir)
    if fn_filter:
        l = [path + f for f in files if f in fn_filter]
    else:
        l = [path + f for f in files if not f.startswith('.')]
    l.sort()
    return l

def is_time(time):
    try:
        return float(time)
    except:
        return False

def find_times(fold=None):
    """ Find time folders in given or current folder
        Returns sorted list of times as strings
    """
    search_folder = (fold if fold else os.getcwd())
    cur_dir = os.walk(search_folder)
    root, dirs, files = next(cur_dir)
    times = [time for time in dirs if is_time(time) is not False]
    times.sort()
    return times

def dataframe_to_foam(fullname, ftype, dataframe, boundaries):
    """ writes an OpenFOAM field file from given dataframe """
    with open(fullname, 'w', encoding='utf-8') as f:
        fname = fullname.split('/')[-1]
        time = fullname.split('/')[-2]
        print("writing %s : %s" % (time, fname))
        f.write(FOAM_HEADER.format(ftype.capitalize(), time, fname, ftype))
        f.write(str(len(dataframe)))
        f.write("\n(\n")
        out=""
        for count, row in dataframe.iterrows():
            if len(row) == 1:
                out = "{}\n".format(row[0])
            elif len(row) == 3:
                out = "({} {} {})\n".format(row[0], row[1], row[2])
            elif len(row) == 6:
                out = "({} {} {} {} {} {})\n".format(row[0], row[1], row[2], row[3], row[4], row[5])
            f.write(out)
        f.write(");\n")
        f.write("boundaryField {\n")
        for bound in boundaries:
            f.write(bound)
            f.write("\t{type zeroGradient;}\n")
        f.write("}")
        f.write("\n// ************************************************************************* //")

class Origins():
    """ Class to manage fields to file relation and store hashes

        dct = {'hash':34jd
               0.0:{'hash':234s                     #time
                    'centreline':{'hash':94143e     #loc
                                  'U':filename,3424}
                    }
              }
    """
    from collections import defaultdict
    def __init__(self):
       self.dct = defaultdict(dict)

    @classmethod
    def from_dict(cls, dct):
        pass

    def to_dict(self):
        pass

    def insert(self, time, loc, field, filename, fieldhash):
        try:
            self.dct[time][loc][field] = filename, fieldhash
        except:
            self.dct[time].update({loc:{field: (filename, fieldhash)}})

    def update_hashes(self):
        for time_key, time in self.dct.iteritems():
            if time_key == "hash":
                continue
            for loc_key, loc in time.iteritems():
                if loc_key == "hash":
                    continue
                self.dct[time_key][loc_key]["hash"] = sum(
                    [field[1] for key,field in loc.iteritems() if key != "hash"]
                )
            self.dct[time_key]["hash"] = sum(
                    [field["hash"] for key,field in time.iteritems() if key != "hash"]
            )
        self.dct["hash"] = sum([field["hash"] for key,field in self.dct.iteritems()
                                    if key != "hash"]
        )

    def hashes(self):
        """ generator """
        # self.update_hashes()
        for time_key, time in self.dct.iteritems():
            if time_key == "hash":
                continue
            for loc_key, loc in time.iteritems():
                if loc_key == "hash":
                    continue
                for field, item in loc.iteritems():
                    if field == "hash":
                        continue
                    fn, field_hash = item
                    yield ((time_key, self.dct["hash"]),
                           (loc_key,  time["hash"]),
                           (field, loc["hash"]),
                           (fn, field_hash)
                          )

    def find(self, search_hash):
        for time, loc, field, item in self.hashes():
            time_name, time_hash   = time
            loc_name, loc_hash     = loc
            field_name, field_hash = field
            filename, item_hash    = item
            if (search_hash == item_hash):
                return field_name, filename
        else:
            return None,None

class ProgressBar():
    """ A class providing progress bars """

    def __init__(self, n_tot, bins=10):
        #FEATURE: Add timings
        self.tot = float(n_tot)
        self.count = 0.0
        self.cur = 0.0

    def next(self):
        self.count += 1.0
        if self.count/self.tot > self.cur:
            print("#", end="")
            self.cur += 0.1

    def done(self):
        print("[done]")

def strip_time(path, base):
    """ try to extract time from path """
    wo_base = path.replace(base, '')
    wo_proc = re.sub('processor[0-9]?', '', wo_base)
    match = re.search(FPNUMBER, wo_proc)
    if match:
        time = float(match.group())
        return time
    else:
        return 0.0

def import_foam_mesh(path, exclude=None, times_slice=None):
    """ returns a Dataframe containing the raw mesh data """
    mesh_loc = "constant/polyMesh"
    if mesh_loc not in path:
        path = os.path.join(path, mesh_loc)

    fileList = find_datafiles(
            path,
            search="[.\/A-Za-z]*",
            files=['faces', 'points', 'owner', 'neighbour'],
            exclude=exclude,
            times_slice=times_slice,
        )
    if not fileList:
        print("no mesh files found")
        return
    p_bar = ProgressBar(n_tot=sum([len(l) for l in fileList.values()]))
    df = DataFrame()
    from collections import defaultdict
    origins = Origins()
    els = list(fileList.items())
    time, files = els[0]
    df_tmp = dict()
    for fn in files:
        ret = read_data_file(fn, skiplines=1, maxlines=False)
        p_bar.next()
        field_names, x, hashes = ret
        df_tmp[fn] = x
    return df_tmp


def import_foam_folder(
        path,
        search,
        files,
        skiplines=1,
        maxlines=0,
        skiptimes=1,
        exclude=None,
        times_slice=None,
        ):
    """ returns a Dataframe for every file in fileList """
    #import StringIO
    fileList = find_datafiles(
        path, search=search, files=files,
        exclude=exclude, times_slice=times_slice
    )
    if not fileList:
        print("no files found")
        return None, DataFrame()
    p_bar = ProgressBar(n_tot=sum([len(l) for l in fileList.values()]))
    df = DataFrame()
    #df.index = MultiIndex.from_tuples(zip([],[]),names=['Loc',0])
    from collections import defaultdict
    origins = Origins()
    els = list(fileList.items())[::skiptimes]
    for fullpath, files in els:
        time = strip_time(fullpath, path)
        df_tmp = DataFrame()
        for fn in files:
            #ret = read_table(StringIO.StringIO(foam_to_csv(fn)))
            ret = read_data_file(fn, skiplines, maxlines)
            p_bar.next()
            if not ret or ret[1].empty:
                continue
            field_names, x, hashes = ret
            loc = x.index.values[-1][0]
            if df_tmp.empty:
                df_tmp = x
            else:
                try:
                    df_tmp = df_tmp.combine_first(x)
                except Exception as e:
                    print("failed to concat: ",
                            df_tmp, "and", x, "new_loc ",
                            x.index.levels[0][0], " existing_locs ",
                            df_tmp.index.levels[0] )
                    print(e)
            field_names = ([field_names] if not type(field_names) == list else field_names)
            for field in field_names:
                if field == "Pos":
                    continue
                origins.insert(time, loc, field, fn, hashes[field])
        df_tmp['Time'] = time
        if df.empty:
            df = df_tmp
        else:
            df = df.append(df_tmp)
    df.set_index('Time', append=True, inplace=True)
    if not "Loc" in  df.index.names:
        print(df)
        # df = df.reorder_levels(['Time', ])
    else:
        df = df.reorder_levels(['Time', 'Loc', 'Pos'])
    p_bar.done()
    return origins, df

"""
Time Loc        Pos U V
1000 radVel+10  0.1
              2
     radVel+20  0.1
                0.2
"""


def foam_to_csv(fn, ):
    """ helper function for d3.js data conversion
        prints data directly to std:out
    """
    try:
        with open(fn, encoding="utf-8") as f:
            content = f.readlines()
            start, num_entries = if_header_skip(content)
            entries = len(content[start].split())
            for l in content:
                print(re.sub("\t",",",re.sub("[\(\)\\n]","",l)))
    except Exception as e:
        print(e)

def read_boundary_names(fn):
    """ Todo use iterator method to avoid reading complete file """
    with open(fn, encoding="utf-8") as f:
        boundary_names = []
        lines  = reversed(f.readlines())
        for line in lines:
            if "{" in line:
                follower = next(lines)
                if "boundaryField" in follower:
                    return boundary_names
                try:
                     boundary_names.append(follower)
                except:
                    pass
            else:
                pass

def read_data_file(fn, skiplines=1, maxlines=False):
    """  A function to read any foam data files returning data and
         index after header
    """

    # TODO check if sorting the index gives any performance benefits
    # print "opening file {}".format(fn)
    if not os.path.exists(fn):
        print("Can not open file " + fn)
        return None
    try:
        with open(fn, encoding="utf-8") as f:
            field = fn.split('/')[-1]
            content = f.readlines()
            content.append('bla')
            start, num_entries = if_header_skip(content)
            entries = len(content[start].split())
            is_a_vector = (True if entries > 1 else False)
            end = start + num_entries
            # FIXME this fails for eulerian/lagrangian vector fields
            # since no positional entry is produced
            if is_a_vector:
                data = list(map(lambda x: re.sub("[0-9]*\(|\)", '', x).split(),
                            content[start:end:skiplines]))
                loc, names = evaluate_names(fn, entries)
                df = DataFrame(data=data, columns=names)
                if loc:
                    df['Loc'] = loc
                else:
                    df['Loc'] = range(len(df))
                if "Pos" in df:
                    df.set_index('Loc', append=False, inplace=True)
                    df["Pos"] = df["Pos"].astype(float)
                    df.set_index('Pos', append=True, inplace=True)
                else:
                    # if no pos is availible we have either
                    # an eulerian or lagrangian field
                    df.set_index('Loc', append=True, inplace=True)
                    df.index.names = ['Pos', 'Loc']
                    df = df.reorder_levels(['Loc', 'Pos'])
                df = df.astype(float)
                hashes = {}
                for row in df.columns:
                    hashes.update({row: hash_series(df[row])})
                return names, df, hashes
            # DataFile with a single row are seen as Eulerian or Lagrangian fields
            else:
                data = [np.float32(x) for x in content[start:end:skiplines]]
                entries = 1
                df = DataFrame(data=data, columns=[field])
                df['Loc'] = "Field"
                df.set_index('Loc', append=True, inplace=True)
                df.index.names=['Pos', 'Loc']
                df = df.reorder_levels(['Loc', 'Pos'])
                hashes = {field: int(hashlib.md5(str(data).encode('utf-8')).hexdigest(),16)}
                return field, df, hashes
    except Exception as e:
        if DEBUG:
            print("Error processing datafile " + fn)
            print(e)
        return None

def hash_series(series):
    d = series.values
    d.flags.writeable = False #TODO needed?
    s = str(list(d)).encode('utf-8')
    return int(hashlib.md5(s).hexdigest(),16) #NOT


def evaluate_names(fullfilename, num_entries):
    """ Infere field names and Loc from given filename

        Example:
            U -> Field, [u,v,w]
            centreLine_U.xy -> centreLine, [Pos,u,v,w]
    """
    filename = fullfilename.split('/')[-1]
    name = (filename.replace('.dat','') #FIXME use regex here
            .replace('.xy','')
            .replace('UMean','uMean_vMean_wMean')
            .replace('UPrime2Mean','uu_uv_uw_vv_vw_ww')
            .replace('Uc','uc_vc_wc')
            .replace('U','u_v_w')
           )
    fields = name.split('_')
    if num_entries == len(fields):
        pos = "Field"
        if '.dat' in filename or '.xy' in filename:
            pos = fields[0]
            fields[0] = "Pos"
        return pos, fields
    else:
        return "Unknown", [filename + '_' + str(i) for i in range(num_entries)]


def req_file(file_name, requested):
    """ True if file name is list of requested files """
    if requested == True:
        return requested
    else:
        return file_name.split('/')[-1] in requested


def import_logs(folder, search, keys):
    """
        keys = {"ExectionTime": ["ExecTime", "ClockTime"]}

        return a DataFrame

              Loc, Time KeyName1 Keyname2
                1   0.1

                    0.2
                2


    """
    def find_start(log):
        """ Fast forward through file till 'Starting time loop' """
        for i, line in enumerate(log):
            if "Starting time loop" in line:
                return i


    def extract(line, keys):
        """
            returns key and values as list
                "ExecutionTime":[0,1]
        """
        import re
        for key, col_names in keys.items():
            if re.search(key, line):
                return col_names, list(
                        map(float,filter(lambda x:
                        x, re.findall("[0-9]+[.]?[0-9]*[e]?[\-]?[0-9]*", line))))
        return None, None

    fold, dirs, files = next(os.walk(folder))
    logs = [fold + "/" + log for log in files if search in log]
    p_bar = ProgressBar(n_tot = len(logs))
    # Lets make sure that we find Timesteps in the log
    keys.update({"^Time = ": ['Time']})

    for log_number, log_name in enumerate(logs):
        with open(log_name, encoding="utf-8") as log:
            f = log.readlines()
            start = find_start(f)
            dataDict = defaultdict(list)
            df=DataFrame()
            for line in f[start:-1]:
                 col_names, values = extract(line, keys)
                 if not col_names:
                    continue
                 if col_names[0] == 'Time':
                    # a new time step has begun
                    # flush datadict and concat to df
                    # Very slow but, so far the solution
                    # to keep subiterations attached to correct time
                    # FIXME: still needs handling of different length dictionaries
                    df = concat([df, DataFrame(dataDict)])
                    dataDict = defaultdict(list)
                 for i, col in enumerate(col_names):
                    dataDict[col].append(values[i])
        p_bar.next()
        try:
            df.index=range(len(df))
            df.index.names=['Id']
            df['Loc'] = log_number
            df.set_index('Time', append=True, inplace=True)
            df.set_index('Loc', append=True, inplace=True)
            df = df.reorder_levels(['Loc','Time','Id'])
            p_bar.done()
        except Exception as e:
            print(log_name)
            print("failed to process")
            print(e)
            return {}, None
    return {}, DataFrame()



def if_header_skip(content):
    """ go through first lines of file and check if has header
        return start of content and total lines
    """
    def begins_with_int(line):
        try:
            num = int(line)
            return num
        except:
            return False
    first_line = content[0]
    if not first_line.startswith('#') and not first_line.startswith('/*'):
        return 0, -1
    elif first_line.startswith('#'):
        for line_number, line in enumerate(content):
            if not line.startswith('#'):
                return line_number, -line_number-1 #FIXME dont append bla
    elif first_line.startswith('/*'):
        for line_number, line in enumerate(content):
            entries = begins_with_int(line)
            if entries or line_number >= 100:
                return line_number + 2, entries



from __future__ import print_function
from future.builtins import *

from collections import OrderedDict
from itertools import cycle


from .FoamFrame import FoamFrame
from .plot import style as defstyle
from .plot import arangement, compose_styles, colored


import bokeh.plotting as bk


def multiframes(folder, names, reader, **kwargs):
    """ create a collection of cases for which
        only the folder is different """
    return MultiFrame([reader(f, name=n, **kwargs)
                       for (f, n) in zip(folder, names)])


class MultiFrame():
    """ Class for storage of multiple case items
        or faceted data from FoamFrame
    """

    @staticmethod
    def from_dict(input_dict, **kwargs):
        return MultiFrame(
                cases=[FoamFrame.from_dict(d, name=name, **kwargs)
                        for name, d in input_dict.items()]
                )

    def __repr__(self):
        s = "MultiFrame with {} entries:\n".format(len(self.cases))
        s += "\n".join(["{}\n{}:\n{}".format(80*"=", name, c.describe())
                        for name, c in self.cases.items()])
        return s

    def __init__(self, cases=None):
        if type(cases) == list:
            self.cases = OrderedDict([(case.properties.name, case)
                                      for case in cases])
        elif type(cases) == OrderedDict:
            self.cases = cases
        else:
            self.cases = {}

    def __getitem__(self, field):
        return [serie[field] for serie in self.cases.values()]

    def names(self):
        return [name for name in self.cases]

    def select(self, case):
        """ select a specific item """
        return self.cases[case]

    def values(self):
        for case in self.cases.values():
            yield case

    def items(self):
        for name, case in self.cases.items():
            yield name, case

    def insert(self, key, value):
        self.cases[key] = value


    def histogram(self, y, x=None, **kwargs):
        cases = list(self.cases.keys())
        fig = self.cases[cases[0]].histogram(x=x, y=y, **kwargs)
        for c in cases:
            fig = self.cases[c].histogram(x=x, y=y, figure=fig, **kwargs)
        return fig


    def show(self, y, x='Pos', z=False, overlay="Field",
             style=defstyle, filename=None, show=True, **kwargs):
        """ Display single quantity y over multiple cases
            if overlay is set all cases are plotted in to single
            graph """
        # TODO if not overlay, common part should be figure title
        style = (compose_styles(style, []) if isinstance(style, list) else style)
        dashes = cycle([[8, 4], [4, 4], [4, 2], [1, 1]])
        cases = list(self.cases.keys())
        row = self.cases[cases[0]].show(x=x, y=y, overlay=overlay,
                                        legend_prefix=cases[0], style=style,
                                        post_pone_style=True, titles=y, **kwargs)
        for c, d, col in zip(cases[1:], dashes, colored[1:]):
            row = self.cases[c].show(x=x, y=y, overlay=overlay,
                                     legend_prefix=c, style=style,
                                     row=row, post_pone_style=True,
                                     line_dash=d, titles=y, color=col, **kwargs)
        gp = bk.GridPlot(children=style(rows=arangement(list(row.values()))))
        if filename:
            bk.save(gp, filename)
        if show:
            return bk.show(gp)
        else:
            return gp

    # def show_multi(self, ys, locs, x='Pos', style=defstyle, **kwargs):
    #     bk.figure()
    #     rows=[]
    #     ys = (ys if isinstance(ys, list) else [ys])
    #     for i, y in enumerate(ys):
    #         figs = plot.plot_cases(self, y=y, x=x, order=locs,
    #                 legend=True, **kwargs)
    #         rows.append(figs)
    #     return bk.GridPlot(children=style(rows=rows))
    #

    # ----------------------------------------------------------------------
    # Filter methods

    def filter(self, name, index=None, field=None):
        for cname, case in self.cases.items():
            self.cases[cname] = case.filter(name, index, field)
        return self

    def filter_fields(self, name, lower, upper):
        """ filter based on field values

            Examples:

                .filter_fields('T', 1000, 2000)
        """
        return self.filter(name, field=lambda x: lower < x < upper)

    def filter_locations(self, index):
        """ filter based on locations

            Examples:

                .filter_location(Owls.isIn('radial'))
                .filter_location(Owls.isNotIn('radial'))

        """
        return self.filter(name='Loc', index=index)

    def filter_items(self, func):
        """ select items based on filter funtion

            Example .filter_items(lambda ff: "Foo" in ff.locations)
        """
        return MultiFrame(filter(func, self.cases.items()))

    # ----------------------------------------------------------------------
    # Selection methods

    def location(self, loc):
        return MultiFrame([c.location(loc) for _, c in self.cases.items()])

    @property
    def latest(self):
        """ Grouping delegator """
        return MultiFrame([case.latest for cname, case in self.cases.items()])

    def on(self, case, func, **kwargs):
        """
            mf.on('Exp', "location", loc='axis')
        """
        n = getattr(self.cases[case], func)(**kwargs)
        return MultiFrame([(case_ if cname != case else n)
                           for cname, case_ in self.cases.items()])

    # ----------------------------------------------------------------------
    # Grouping methods

    def by_index(self, field, func=None):
        func = (func if func else lambda x: x)
        return self.by(field, func)

    def by_field(self, field, func=None):
        func = (func if func else lambda x: x)
        return self.by(field, func)

    def by_location(self, func=None):
        func = (func if func else lambda x: x)
        return self.by("Loc", func)

    def by_time(self, func=None):
        func = (func if func else lambda x: x)
        return self.by("Time", func)

    def by(self, name, func):
        """ Grouping delegator """
        return MultiFrame([case.by(name, func)
                           for cname, case in self.cases.items()])

    def update_plot_properties(self, field, d):
        """ update plot properties of all casese """

        for _, c in self.cases.items():
                c.properties.plot_properties.insert(field, d)
        return self



#!/usr/bin/python
"""
    Usage: 
        ana.py case2case [--avg] [--div] --times=<times> --vectors=<fields> --scalars=<fields> --tensors=<field> --case=<dir>
"""

from io import *
import os
from docopt import docopt
from pandas import *

def file_names(arguments):
    """ construct file names for read in from cli args """
    for ftype in ['vector', 'scalar', 'tensor']:
        fargs = '--' + ftype + 's'
        if not arguments[fargs]:
            continue
        cwd, fields, times = os.getcwd(), arguments[fargs], arguments['--times']
        fields = fields.split(',')
        for fname in fields:
            this = "{}/{}/{}".format(cwd, times, fname)
            if arguments['--case']:
                that = "{}/{}/{}/{}".format(cwd, arguments['--case'], times, fname)
            else:
                that = None
            ftype = ftype.replace('tensor','symTensor')
            yield ftype, fname, times, this, that

if __name__ == '__main__':
    arguments = docopt(__doc__)
    for ftype, fname, times, ours, theirs in file_names(arguments):
        origins_this, this = read_data_file(ours, 1, 0, {})
        origins_that, that = read_data_file(theirs, 1, 0, {})
        boundaries = read_boundary_names("{}/{}/{}".format(os.getcwd(), times, fname))
        diff, avg = this-that, 0.5*(this+that)
        for name, result in {fname+'diff': diff, fname+'avg': avg}.iteritems():
            fullname = "{}/{}/{}".format(os.getcwd(), times, name)
            dataframe_to_foam(fullname, ftype, result, boundaries)



import os
import subprocess
import sys

from setuptools import setup

#from io import open

package_name = 'Owls'
# Fetch version from git tags, and write to version.py.
# Also, when git is not available (PyPi package), use stored version.py.
version_py = os.path.join(os.path.dirname(__file__), 'Owls/version.py')

try:
    version_git = subprocess.check_output(["git", "describe"], universal_newlines=True).rstrip()
except:
    with open(version_py, 'w') as fh:
        version_git = open(version_py).read().strip().split('=')[-1].replace('"','')

version_msg = "# Do not edit this file, pipeline versioning is governed by git tags"
with open(version_py, 'w') as fh:
    fh.write(version_msg + os.linesep + "__version__='{}'".format(version_git))

# Get the version from Owls/version.py without importing the package
# exec(compile(open(package_name + '/version.py').read(),
#              package_name + '/version.py', 'exec'))

packages = [
    'future',
    'numpy',
    'pandas==0.17.1',
    'functional',
    'bokeh==0.10.0',
    ]

if '--with-matplotlib' in sys.argv:
    packages.extend('matplotlib')

config = {
    'author': 'Gregor Olenik',
    'author_email': 'itvgo@itv.uni-stuttgart.de',
    'description': 'A simple library for reading and processing OpenFOAM data',
    'license': 'BSD',
    'version': version_git,
    'packages': [package_name],
    'install_requires': packages,
    'data_files': [('scripts', ['scripts/snap.py'])],
    'name': package_name,
    'zip_safe': False
}


setup(**config)



"""
Launchers for openfoam utilities

"""
import os
import subprocess

def execute(cmd):
    return subprocess.check_call(cmd, shell=True) == 0     

def execute_in_path(path, cmd, func=execute):
    print(os.getcwd())
    old_dir = os.getcwd()
    os.chdir(path)
    try:
        ret = func(cmd)
    except:
        ret = False
    os.chdir(old_dir)
    return ret

class genericHook(object):

    def __init__(self, path, cmd):
        self.path = path
        self.cmd = cmd
    
    def execute(self):
        return execute_in_path(self.path, self.cmd)

class cellCentres(genericHook):

    def __init__(self, path):
        genericHook.__init__(self, path, "writeCellCentres")
        
class decompose(genericHook):

    def __init__(self, path, latest=False):
        cmd = "decomposePar"
        if latest:
            cmd += " -latestTime"
        genericHook.__init__(self, path, cmd)

class reconstruct(genericHook):

    def __init__(self, path, latest=False):
        cmd = "reconstructPar"
        if latest:
            cmd += " -latestTime"
        genericHook.__init__(self, path, cmd)

class sample(genericHook):

    def __init__(self, path, latest=False):
        cmd = "sample"
        if latest:
            cmd += " -latestTime"
        genericHook.__init__(self, path, cmd)
